<#--
MIT License

Copyright (c) 2020 GitHub user u/garronej

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
-->

<#--
Code derived from Keycloakify (https://github.com/keycloakify/keycloakify/blob/main/src/bin/keycloakify/generateFtl/kcContextDeclarationTemplate.ftl)
-->

<#function toJsDeclarationString object path>
    <#local isHash = -1>
    <#attempt>
        <#local isHash = object?is_hash || object?is_hash_ex>
    <#recover>
        <#return "ABORT: Can't evaluate if " + path?join(".") + " is a hash">
    </#attempt>

    <#if isHash>
        <#if path?size gt 10>
            <#return "ABORT: Too many recursive calls, path: " + path?join(".")>
        </#if>
        <#local keys = -1>

        <#attempt>
            <#local keys = object?keys>
        <#recover>
            <#return "ABORT: We can't list keys on object">
        </#attempt>

        <#local outSeq = []>

        <#list keys as key>
            <#if ["class","declaredConstructors","superclass","declaringClass" ]?seq_contains(key) >
                <#continue>
            </#if>

            <#if (
                    areSamePath(path, ["url"]) &&
                    ["loginUpdatePasswordUrl", "loginUpdateProfileUrl", "loginUsernameReminderUrl", "loginUpdateTotpUrl"]?seq_contains(key)
                ) || (
                    key == "updateProfileCtx" && 
                    areSamePath(path, [])
                ) || (
                    <#-- https://github.com/keycloakify/keycloakify/pull/65#issuecomment-991896344 (reports with saml-post-form.ftl) -->
                    <#-- https://github.com/keycloakify/keycloakify/issues/91#issue-1212319466 (reports with error.ftl and Kc18) -->
                    <#-- https://github.com/keycloakify/keycloakify/issues/109#issuecomment-1134610163 -->
                    <#-- https://github.com/keycloakify/keycloakify/issues/357 -->
                    <#-- https://github.com/keycloakify/keycloakify/discussions/406#discussioncomment-7514787 -->
                    key == "loginAction" && 
                    areSamePath(path, ["url"]) && 
                    ["saml-post-form.ftl", "error.ftl", "info.ftl", "login-oauth-grant.ftl", "logout-confirm.ftl", "login-oauth2-device-verify-user-code.ftl"]?seq_contains(pageId) &&
                    !(auth?has_content && auth.showTryAnotherWayLink())
                ) || (
                    <#-- https://github.com/keycloakify/keycloakify/issues/362 -->
                    ["secretData", "value"]?seq_contains(key) && 
                    areSamePath(path, [ "totp", "otpCredentials", "*" ])
                ) || (
                    ["contextData", "idpConfig", "idp", "authenticationSession"]?seq_contains(key) &&
                    areSamePath(path, ["brokerContext"]) &&
                    ["login-idp-link-confirm.ftl", "login-idp-link-email.ftl" ]?seq_contains(pageId)
                ) || (
                    key == "identityProviderBrokerCtx" && 
                    areSamePath(path, []) &&
                    ["login-idp-link-confirm.ftl", "login-idp-link-email.ftl" ]?seq_contains(pageId)
                ) ||  (
                    ["masterAdminClient", "delegateForUpdate", "defaultRole", "smtpConfig"]?seq_contains(key) &&
                    areSamePath(path, ["realm"])
                ) || (
                    areSamePath(path, ["realm"]) &&
                    !["name", "displayName", "displayNameHtml", "internationalizationEnabled", "registrationEmailAsUsername" ]?seq_contains(key)
                ) || (
                    ( 
                        key == "realm" || 
                        key == "container" 
                    ) &&
                    isSubpath(path, ["applications", "applications"])
                ) || (
                    key == "delegateForUpdate" &&
                    areSamePath(path, ["user"])
                ) || (
                    <#-- Security audit forwarded by Garth (Gmail) -->
                    key == "saml.signing.private.key" &&
                    areSamePath(path, ["client", "attributes"])
                ) || (
                    <#-- See: https://github.com/keycloakify/keycloakify/issues/534 -->
                    key == "password" &&
                    areSamePath(path, ["login"])
                ) || (
                    <#-- Remove realmAttributes added by https://github.com/jcputney/keycloak-theme-additional-info-extension for peace of mind. -->
                    key == "realmAttributes" &&
                    areSamePath(path, [])
                ) || (
                    <#-- attributesByName adds a lot of noise to the output and is not needed, we already have profile.attributes -->
                    key == "attributesByName" &&
                    areSamePath(path, ["profile"]) 
                ) || (
                    <#-- We already have the attributes in profile speedup the rendering by filtering it out from the register object -->
                    (key == "attributes" || key == "attributesByName") &&
                    areSamePath(path, ["register"])
                ) || (
                    areSamePath(path, ["properties"]) &&
                    (
                        key?starts_with("kc") || 
                        key == "locales" || 
                        key == "import" || 
                        key == "parent" || 
                        key == "meta" ||
                        key == "stylesCommon" ||
                        key == "styles" ||
                        key == "accountResourceProvider"
                    )
                ) || (
                    key == "execution" &&
                    areSamePath(path, [])
                ) || (
                    key == "entity" &&
                    areSamePath(path, ["user"])
                ) || (
                    key == "attributes" &&
                    areSamePath(path, ["realm"])
                ) || (
                    areSamePath(path, ["realm"]) &&
                    ![ 
                        "name",
                        "registrationEmailAsUsername",
                        "editUsernameAllowed",
                        "isInternationalizationEnabled",
                        "identityFederationEnabled",
                        "userManagedAccessAllowed"
                    ]?seq_contains(key)
                ) || (
                    ["flowContext", "session", "realm"]?seq_contains(key) &&
                    areSamePath(path, ["social"])
                )
            >
                <#-- <#local outSeq += ["/*" + path?join(".") + "." + key + " excluded*/"]> -->
                <#continue>
            </#if>

            <#-- https://github.com/keycloakify/keycloakify/discussions/406 -->
            <#if (
                key == "attemptedUsername" &&
                areSamePath(path, ["auth"]) &&
                [ 
                    "register.ftl", "terms.ftl", "info.ftl", "login.ftl", 
                    "login-update-password.ftl", "login-oauth2-device-verify-user-code.ftl"
                ]?seq_contains(pageId)
            )>
                <#attempt>
                    <#-- https://github.com/keycloak/keycloak/blob/3a2bf0c04bcde185e497aaa32d0bb7ab7520cf4a/themes/src/main/resources/theme/base/login/template.ftl#L63 -->
                    <#if !(auth?has_content && auth.showUsername() && !auth.showResetCredentials())>
                        <#local outSeq += ["/*" + path?join(".") + "." + key + " excluded*/"]>
                        <#continue>
                    </#if>
                <#recover>
                    <#local outSeq += ["/*Accessing attemptedUsername throwed an exception */"]>
                </#attempt>
            </#if>

            {{userDefinedExclusions}}

            <#attempt>
                <#if !object[key]??>
                    <#continue>
                </#if>
            <#recover>
                <#local outSeq += ["/*Couldn't test if '" + key + "' is available on this object*/"]>
                <#continue>
            </#attempt>

            <#local propertyValue = -1>

            <#attempt>
                <#local propertyValue = object[key]>
            <#recover>
                <#local outSeq += ["/*Couldn't dereference '" + key + "' on this object*/"]>
                <#continue>
            </#attempt>

            <#local recOut = toJsDeclarationString(propertyValue, path + [ key ])>

            <#if recOut?starts_with("ABORT:")>

                <#local errorMessage = recOut?remove_beginning("ABORT:")>

                <#if errorMessage != " It's a method" >
                    <#local outSeq += ["/*" + key + ": " + errorMessage + "*/"]>
                </#if>

                <#continue>
            </#if>

            <#local outSeq +=  ['"' + key + '": ' + recOut + ","]>

        </#list>

        <#return (["{"] + outSeq?map(str -> ""?right_pad(4 * (path?size + 1)) + str) + [ ""?right_pad(4 * path?size) + "}"])?join("\n")>

    </#if>

    <#local isMethod = -1>
    <#attempt>
        <#local isMethod = object?is_method>
    <#recover>
        <#return "ABORT: Can't test if it'sa method.">
    </#attempt>

    <#if isMethod>

        <#if areSamePath(path, ["auth", "showUsername"])>
            <#attempt>
                <#return auth.showUsername()?c>
            <#recover>
                <#return "ABORT: Couldn't evaluate auth.showUsername()">
            </#attempt>
        </#if>

        <#if areSamePath(path, ["auth", "showResetCredentials"])>
            <#attempt>
                <#return auth.showResetCredentials()?c>
            <#recover>
                <#return "ABORT: Couldn't evaluate auth.showResetCredentials()">
            </#attempt>
        </#if>

        <#if areSamePath(path, ["auth", "showTryAnotherWayLink"])>
            <#attempt>
                <#return auth.showTryAnotherWayLink()?c>
            <#recover>
                <#return "ABORT: Couldn't evaluate auth.showTryAnotherWayLink()">
            </#attempt>
        </#if>

        <#if areSamePath(path, ["url", "getLogoutUrl"])>
            <#local returnValue = -1>
            <#attempt>
                <#local returnValue = url.getLogoutUrl()>
            <#recover>
                <#return "ABORT: Couldn't evaluate url.getLogoutUrl()">
            </#attempt>
            <#return 'function(){ return "' + returnValue + '"; }'>
        </#if>

        <#if areSamePath(path, ["totp", "policy", "getAlgorithmKey"])>
            <#local returnValue = "error">
            <#if mode?? && mode = "manual">
                <#attempt>
                    <#local returnValue = totp.policy.getAlgorithmKey()>
                <#recover>
                    <#return "ABORT: Couldn't evaluate totp.policy.getAlgorithmKey()">
                </#attempt>
            </#if>
            <#return 'function(){ return "' + returnValue + '"; }'>
        </#if>

        <#if profile?? && profile.attributes??>
            <#list profile.attributes as attribute>
                <#if fieldNames?seq_contains(attribute.name)>
                    <#continue>
                </#if>
                <#assign fieldNames += [attribute.name]>
            </#list>
        </#if>

        <#return "ABORT: It's a method">
    </#if>

    <#local isBoolean = -1>
    <#attempt>
        <#local isBoolean = object?is_boolean>
    <#recover>
        <#return "ABORT: Can't test if it's a boolean">
    </#attempt>

    <#if isBoolean>
        <#return object?c>
    </#if>

    <#local isEnumerable = -1>
    <#attempt>
        <#local isEnumerable = object?is_enumerable>
    <#recover>
        <#return "ABORT: Can't test if it's an enumerable">
    </#attempt>


    <#if isEnumerable>

        <#local outSeq = []>

        <#local i = 0>

        <#list object as array_item>

            <#if !array_item??>
                <#local outSeq += ["null,"]>
                <#continue>
            </#if>

            <#local recOut = toJsDeclarationString(array_item, path + [ i ])>

            <#local i = i + 1>

            <#if recOut?starts_with("ABORT:")>

                <#local errorMessage = recOut?remove_beginning("ABORT:")>

                <#if errorMessage != " It's a method" >
                    <#local outSeq += ["/*" + i?string + ": " + errorMessage + "*/"]>
                </#if>

                <#continue>
            </#if>

            <#local outSeq += [recOut + ","]>

        </#list>

        <#return (["["] + outSeq?map(str -> ""?right_pad(4 * (path?size + 1)) + str) + [ ""?right_pad(4 * path?size) + "]"])?join("\n")>

    </#if>

    <#local isDate = -1>
    <#attempt>
        <#local isDate = object?is_date_like>
    <#recover>
        <#return "ABORT: Can't test if it's a date">
    </#attempt>

    <#if isDate>
        <#return '"' + object?datetime?iso_utc + '"'>
    </#if>

    <#local isNumber = -1>
    <#attempt>
        <#local isNumber = object?is_number>
    <#recover>
        <#return "ABORT: Can't test if it's a number">
    </#attempt>

    <#if isNumber>
        <#return object?c>
    </#if>

    <#local isString = -1>
    <#attempt>
        <#local isString = object?is_string>
    <#recover>
        <#return "ABORT: Can't test if it's a string">
    </#attempt>

    <#attempt>
        <#return '"' + object?js_string + '"'>;
    <#recover>
    </#attempt>

    <#return "ABORT: Couldn't convert into string non hash, non method, non boolean, non number, non enumerable object">

</#function>
<#function isSubpath path searchedPath>

    <#if path?size < searchedPath?size>
        <#return false>
    </#if>

    <#local i=0>

    <#list path as property>

        <#if i == searchedPath?size >
            <#continue>
        </#if>

        <#local searchedProperty=searchedPath[i]>

        <#local i+= 1>

        <#if searchedProperty?is_string && searchedProperty == "*">
            <#continue>
        </#if>

        <#if searchedProperty?is_string && !property?is_string>
            <#return false>
        </#if>

        <#if searchedProperty?is_number && !property?is_number>
            <#return false>
        </#if>

        <#if searchedProperty?string != property?string>
            <#return false>
        </#if>

    </#list>

    <#return true>

</#function>

<#function areSamePath path searchedPath>
    <#return path?size == searchedPath?size && isSubpath(path, searchedPath)>
</#function>

<#function removeBrackets str>
    <#if str?starts_with("${") && str?ends_with("}")>
        <#return str[2..(str?length-2)]>
    <#else>
        <#return str>
    </#if>
</#function>

<#macro baseLayout pageId="template.ftl">
<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta name="robots" content="noindex, nofollow">
    <link rel="icon" href="${url.resourcesPath}/img/favicon.ico">
    <title>${msg("loginTitle",(realm.displayName!''))}</title>
    <style>
      body {
        background-image: url("${url.resourcesPath}/img/background.jpg");
        background-repeat: no-repeat;
        background-position: center;
        background-size: cover;
      }
    </style>
    <script>
      window.kcContext = ${toJsDeclarationString(.data_model, [])?no_esc};
    </script>
    <#nested "head">
</head>
<body>
    <div id="app"></div>
    <#nested "scripts">
</body>
</html>
</#macro>